from __future__ import annotations

import copy
import dataclasses
import functools
import inspect
import sys
import traceback
from functools import reduce
from pathlib import PurePath
from types import AsyncGeneratorType
from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    Iterable,
    List,
    NoReturn,
    Optional,
    Type,
    TypeVar,
    Union,
)

if sys.version_info[:2] < (3, 8):
    from typing_extensions import Literal
else:
    from typing import Literal

from baize.asgi import Receive, Scope, Send
from baize.datastructures import URL
from baize.typing import ASGIApp

from .debug import DebugMiddleware
from .exceptions import ErrorView, ExceptionContextManager, HTTPException
from .requests import HttpRequest, WebSocket, request_var, websocket_var
from .responses import (
    FileResponse,
    HttpResponse,
    JSONResponse,
    PlainTextResponse,
    RedirectResponse,
    SendEventResponse,
    convert_response,
)
from .routing import BaseRoute, NoMatchFound, Router
from .templates import BaseTemplates
from .utils import F, ImmutableAttribute, State

NoArgumentCallable = Union[Callable[[], Any], Callable[[], Awaitable[Any]]]

T_NoArgumentCallable = TypeVar("T_NoArgumentCallable", bound=NoArgumentCallable)


@dataclasses.dataclass
class Lifespan:
    on_startup: List[NoArgumentCallable] = dataclasses.field(default_factory=list)
    on_shutdown: List[NoArgumentCallable] = dataclasses.field(default_factory=list)

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        """
        Handle ASGI lifespan messages, which allows us to manage application
        startup and shutdown events.
        """
        message = await receive()
        assert message["type"] == "lifespan.startup"
        try:
            for handler in self.on_startup:
                result = handler()
                if inspect.isawaitable(result):
                    await result
        except BaseException:
            msg = traceback.format_exc()
            await send({"type": "lifespan.startup.failed", "message": msg})
            raise
        await send({"type": "lifespan.startup.complete"})

        message = await receive()
        assert message["type"] == "lifespan.shutdown"
        try:
            for handler in self.on_shutdown:
                result = handler()
                if inspect.isawaitable(result):
                    await result
        except BaseException:
            msg = traceback.format_exc()
            await send({"type": "lifespan.shutdown.failed", "message": msg})
            raise
        await send({"type": "lifespan.shutdown.complete"})


@dataclasses.dataclass
class FactoryClass:
    http: Type[HttpRequest] = HttpRequest
    websocket: Type[WebSocket] = WebSocket


class Index:
    debug: ImmutableAttribute[bool] = ImmutableAttribute()
    state: ImmutableAttribute[State] = ImmutableAttribute()

    def __init__(
        self,
        *,
        debug: bool = False,
        templates: Optional[BaseTemplates] = None,
        on_startup: List[NoArgumentCallable] = [],
        on_shutdown: List[NoArgumentCallable] = [],
        routes: Iterable[BaseRoute] = [],
        exception_handlers: Dict[int | Type[Exception], ErrorView] = {},
        factory_class: FactoryClass = FactoryClass(),
        response_converters: Dict[type, Callable[..., HttpResponse]] = {},
    ) -> None:
        self.debug = debug
        self.state = State()
        self.response_converter = create_response_converter(response_converters)
        self.should_exit = False
        self.factory_class = factory_class
        self.templates = templates
        self.router = Router(routes)
        self.lifespan = Lifespan(copy.copy(on_startup), copy.copy(on_shutdown))
        self.exception_contextmanager = ExceptionContextManager(exception_handlers)
        self._asgi_middlewares: List[F] = []
        self.app_with_middlewares = self.build_app_with_middlewares()

        if self.debug:
            self.add_middleware(DebugMiddleware)

    def build_app_with_middlewares(self) -> ASGIApp:
        return reduce(lambda a, m: m(a), reversed(self._asgi_middlewares), self.app)

    def add_middleware(self, middleware_class: type, **options: Any) -> None:
        """
        Add ASGI middleware
        """
        self._asgi_middlewares.append(F(middleware_class, **options))
        self.app_with_middlewares = self.build_app_with_middlewares()

    def add_exception_handler(
        self, exc_class_or_status_code: int | Type[Exception], handler: ErrorView
    ) -> None:
        self.exception_contextmanager.add_exception_handler(
            exc_class_or_status_code, handler
        )

    def exception_handler(
        self, exc_class_or_status_code: int | Type[Exception]
    ) -> Callable[[ErrorView], ErrorView]:
        def decorator(func: ErrorView) -> ErrorView:
            self.add_exception_handler(exc_class_or_status_code, func)
            return func

        return decorator

    def on_startup(self, func: T_NoArgumentCallable) -> T_NoArgumentCallable:
        self.lifespan.on_startup.append(func)
        return func

    def on_shutdown(self, func: T_NoArgumentCallable) -> T_NoArgumentCallable:
        self.lifespan.on_shutdown.append(func)
        return func

    async def app(self, scope: Scope, receive: Receive, send: Send) -> None:
        scope_type: Literal["lifespan", "http", "websocket"] = scope["type"]
        return await getattr(self, scope_type)(scope, receive, send)

    async def http(self, scope: Scope, receive: Receive, send: Send) -> None:
        request = self.factory_class.http(scope, receive, send)
        token = request_var.set(request)
        try:
            async with self.exception_contextmanager:
                try:
                    path_params, handler = self.router.search("http", request["path"])
                    request["path_params"] = path_params
                    response = convert_response(await handler())
                except NoMatchFound:
                    raise HTTPException(404)
                else:
                    return await response(scope, receive, send)
        finally:
            request_var.reset(token)

    async def websocket(self, scope: Scope, receive: Receive, send: Send) -> None:
        websocket = self.factory_class.websocket(scope, receive, send)
        token = websocket_var.set(websocket)
        try:
            path_params, handler = self.router.search("websocket", websocket["path"])
            websocket["path_params"] = path_params
        except NoMatchFound:
            return await websocket.close(1001)
        else:
            return await handler()
        finally:
            websocket_var.reset(token)

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        scope["app"] = self

        await self.app_with_middlewares(scope, receive, send)


def create_response_converter(
    converters: Dict[type, Callable[..., HttpResponse]]
) -> functools._SingleDispatchCallable[HttpResponse]:
    """
    Create a converter for convert response.
    """

    @functools.singledispatch
    def response_converter(*args: Any) -> HttpResponse:
        raise TypeError(
            f"Cannot find response_converter handler for this type: {type(args[0])}"
        )

    def _none(ret: Type[None]) -> NoReturn:
        raise TypeError(
            "Get 'None'. Maybe you need to add a return statement to the function."
        )

    response_converter.register(type(None), _none)
    response_converter.register(HttpResponse, lambda x: x)
    response_converter.register(
        dict,
        lambda content, status=200, headers=None: JSONResponse(
            content, status, headers
        ),
    )
    response_converter.register(
        list,
        lambda content, status=200, headers=None: JSONResponse(
            content, status, headers
        ),
    )
    response_converter.register(
        tuple,
        lambda content, status=200, headers=None: JSONResponse(
            content, status, headers
        ),
    )
    response_converter.register(
        bytes,
        lambda content, status=200, headers=None: PlainTextResponse(
            content, status, headers
        ),
    )
    response_converter.register(
        str,
        lambda content, status=200, headers=None: PlainTextResponse(
            content, status, headers
        ),
    )
    response_converter.register(
        AsyncGeneratorType,
        lambda content, status=200, headers=None: SendEventResponse(
            content, status, headers
        ),
    )
    response_converter.register(
        PurePath,
        lambda filepath, download_name=None: FileResponse(
            str(filepath), download_name=download_name
        ),
    )
    response_converter.register(
        URL,
        lambda url, status=307, headers=None: RedirectResponse(url, status, headers),
    )

    for type_, converter in converters.items():
        response_converter.register(type_, converter)

    return response_converter



from __future__ import annotations

import functools
import json
import sys
from inspect import isfunction
from typing import TYPE_CHECKING, Any, Callable, Generator, List, TypeVar
from typing import cast as typing_cast

if sys.version_info[:2] < (3, 8):
    from typing_extensions import Literal
else:
    from typing import Literal

from baize.asgi import Message

from .requests import request, websocket
from .responses import HttpResponse

FuncView = TypeVar("FuncView", bound=Callable)


def required_method(method: str) -> Callable[[FuncView], FuncView]:
    """
    Set the acceptable request method of the function
    """
    allow_methods = {"HEAD", "GET"} if method == "GET" else {method}
    headers = {"Allow": ", ".join(allow_methods)}

    def decorator(function: FuncView) -> FuncView:
        if not isfunction(function):
            raise TypeError("`required_method` can only decorate function")

        @functools.wraps(function)
        async def wrapper(*args, **kwargs):
            if request.method in allow_methods:
                return await function(*args, **kwargs)
            elif request.method == "OPTIONS":
                return HttpResponse(headers=headers)
            else:
                return HttpResponse(status_code=405, headers=headers)

        setattr(wrapper, "__method__", method.upper())
        return typing_cast(FuncView, wrapper)

    return decorator


class HttpView:
    HTTP_METHOD_NAMES = [
        "get",
        "post",
        "put",
        "patch",
        "delete",
        "head",
        "options",
        "trace",
    ]

    if TYPE_CHECKING:
        __methods__: List[str]

    def __init_subclass__(cls) -> None:
        cls.__methods__ = [m.upper() for m in cls.HTTP_METHOD_NAMES if hasattr(cls, m)]

    def __await__(self) -> Generator[None, None, Any]:
        return self.__impl__().__await__()

    async def __impl__(self) -> Any:
        handler = getattr(self, request.method.lower(), self.http_method_not_allowed)
        return await handler()

    async def http_method_not_allowed(self) -> HttpResponse:
        return HttpResponse(
            status_code=405, headers={"Allow": ", ".join(self.__methods__)}
        )

    async def options(self) -> HttpResponse:
        """Handle responding to requests for the OPTIONS HTTP verb."""
        return HttpResponse(headers={"Allow": ", ".join(self.__methods__)})


class SocketView:
    encoding: Literal["anystr", "text", "bytes", "json"] = "anystr"

    def __await__(self):
        return self.__impl__().__await__()

    async def __impl__(self) -> None:
        try:
            close_code = 1000
            await self.on_connect()
            while True:
                message = await websocket.receive()
                if message["type"] == "websocket.receive":
                    data = await self.decode(message)
                    await self.on_receive(data)
                elif message["type"] == "websocket.disconnect":
                    close_code = int(message.get("code", 1000))
                    break
        except Exception as exc:
            close_code = 1011
            raise exc
        finally:
            await self.on_disconnect(close_code)

    async def decode(self, message: Message) -> Any:
        if self.encoding == "text":
            if "text" not in message:
                await websocket.close(code=1003)
                raise RuntimeError("Expected text websocket messages, but got bytes")
            return message["text"]

        if self.encoding == "bytes":
            if "bytes" not in message:
                await websocket.close(code=1003)
                raise RuntimeError("Expected bytes websocket messages, but got text")
            return message["bytes"]

        if self.encoding == "json":
            if message.get("text") is not None:
                text = message["text"]
            else:
                text = message["bytes"].decode("utf-8")

            try:
                return json.loads(text)
            except json.decoder.JSONDecodeError:
                await websocket.close(code=1003)
                raise RuntimeError("Malformed JSON data received.")

        return message["text"] if message.get("text") else message["bytes"]

    async def on_connect(self) -> None:
        """Override to handle an incoming websocket connection"""
        await websocket.accept()

    async def on_receive(self, data: Any) -> None:
        """Override to handle an incoming websocket message"""

    async def on_disconnect(self, close_code: int) -> None:
        """Override to handle a disconnecting websocket"""
        await websocket.close(code=close_code)



from __future__ import annotations

import asyncio
import contextvars
import functools
from typing import Any, AsyncGenerator, Awaitable, Callable, Iterator, TypeVar, Union

from indexpy.utils.inspect import is_async_gen_callable

from .utils import is_coroutine_callable, is_gen_callable

T = TypeVar("T")


async def run_in_threadpool(func: Callable[..., T], *args: Any, **kwargs: Any) -> T:
    loop = asyncio.get_running_loop()
    child = functools.partial(func, *args, **kwargs)
    context = contextvars.copy_context()
    return await loop.run_in_executor(None, context.run, child)


class _StopIteration(Exception):
    pass


def _next(iterator: Iterator[T]) -> T:
    try:
        return next(iterator)
    except StopIteration:
        raise _StopIteration


async def iterate_in_threadpool(iterator: Iterator[T]) -> AsyncGenerator[T, None]:
    while True:
        try:
            yield await run_in_threadpool(_next, iterator)
        except _StopIteration:
            break


def always_async(
    func: Callable[..., T],
) -> Union[Callable[..., AsyncGenerator[T, None]], Callable[..., Awaitable[T]]]:
    if is_coroutine_callable(func) or is_async_gen_callable(func):
        return func

    if is_gen_callable(func):

        @functools.wraps(func)
        async def wrapper(*args, **kwargs):
            generator = func(*args, **kwargs)
            async for item in iterate_in_threadpool(generator):
                yield item

    else:

        @functools.wraps(func)
        async def wrapper(*args, **kwargs):
            return await run_in_threadpool(func, *args, **kwargs)

    return wrapper



import importlib
import os
from types import ModuleType
from typing import Any, Optional


class ImportFromStringError(Exception):
    pass


def import_from_string(import_str: str) -> Any:
    module_str, _, attrs_str = import_str.partition(":")
    if not module_str or not attrs_str:
        message = (
            'Import string "{import_str}" must be in format "<module>:<attribute>".'
        )
        raise ImportFromStringError(message.format(import_str=import_str))

    try:
        module = importlib.import_module(module_str)
    except ImportError as exc:
        if exc.name != module_str:
            raise exc from None
        message = 'Could not import module "{module_str}".'
        raise ImportFromStringError(message.format(module_str=module_str))

    instance = module
    try:
        for attr_str in attrs_str.split("."):
            instance = getattr(instance, attr_str)
    except AttributeError:
        message = 'Attribute "{attrs_str}" not found in module "{module_str}".'
        raise ImportFromStringError(
            message.format(attrs_str=attrs_str, module_str=module_str)
        )

    return instance


def import_module(name: str) -> Optional[ModuleType]:
    """
    try importlib.import_module, nothing to do when module not be found.
    """
    if os.path.exists(os.path.join(os.getcwd(), name + ".py")) or os.path.exists(
        os.path.join(os.getcwd(), name, "__init__.py")
    ):
        return importlib.import_module(name)
    return None  # nothing to do when module not be found



from __future__ import annotations

import os
import signal
import subprocess
import sys
import time
from multiprocessing import cpu_count
from typing import List, Union

import click

from .__version__ import __version__
from .routing.commands import display_urls
from .utils import F, import_from_string, import_module


def execute(command: Union[List[str], str]) -> int:
    if isinstance(command, str):
        command = command.split(" ")

    click.echo("Execute command: ", nl=False)
    click.secho(" ".join(command), fg="green")

    process = subprocess.Popen(command, shell=False)

    def sigint_handler(signo, frame):
        process.terminate()
        process.wait()

    signal.signal(signal.SIGTERM, sigint_handler)

    while process.poll() is None:
        time.sleep(1)

    return process.returncode


@click.group(help=f"Index.py {__version__}")
def index_cli():
    pass


try:
    import hypercorn
except ImportError:
    pass
else:

    @click.command(help="use hypercorn to run Index.py application")
    @click.option(
        "--bind",
        default="127.0.0.1:4190",
        show_default=True,
        help="A string of the form: HOST:PORT, unix:PATH, fd://FD.",
    )
    @click.option(
        "--log-level",
        type=click.Choice(["critical", "error", "warning", "info", "debug"]),
        default="info",
        show_default=True,
    )
    @click.option(
        "--worker-class",
        "-k",
        default="asyncio",
        type=click.Choice(["asyncio", "uvloop", "trio"]),
        show_choices=True,
        show_default=True,
    )
    @click.option(
        "--configuration",
        "-c",
        type=click.Path(exists=True, file_okay=True, dir_okay=False, readable=True),
    )
    @click.argument("application")
    def hypercorn_cli(
        worker_class: str,
        configuration: str,
        application: str,
        bind: str,
        log_level: str,
    ):
        sys.path.insert(0, os.getcwd())
        asgi_app = import_from_string(application)
        config = hypercorn.Config()
        if configuration is not None:
            if configuration.endswith(".py"):
                config.from_pyfile(configuration)
            elif configuration.endswith(".toml"):
                config.from_toml(configuration)
            else:
                click.secho(
                    "Please use configuration file path endswith `.py` or `.toml`.",
                    fg="red",
                )
                raise SystemExit(1)
        config.bind = [bind]
        config.loglevel = log_level.upper()
        config.worker_class = worker_class

        create_signal_handle = lambda shutdown_event: lambda sig, frame: (
            setattr(asgi_app, "should_exit", True),  # type: ignore
            shutdown_event.set(),
        )
        if worker_class == "uvloop":
            import uvloop

            uvloop.install()
        if worker_class in ("asyncio", "uvloop"):
            import asyncio

            from hypercorn.asyncio import serve

            loop = asyncio.get_event_loop()

            shutdown_event = asyncio.Event(loop=loop)
            for sig in {signal.SIGINT, signal.SIGTERM}:
                signal.signal(sig, create_signal_handle(shutdown_event))

            loop.run_until_complete(
                serve(asgi_app, config, shutdown_trigger=shutdown_event.wait)  # type: ignore
            )
        else:
            import trio
            from hypercorn.trio import serve  # type: ignore

            shutdown_event = trio.Event()
            for sig in {signal.SIGINT, signal.SIGTERM}:
                signal.signal(sig, create_signal_handle(shutdown_event))

            trio.run(serve(asgi_app, config, shutdown_trigger=shutdown_event.wait))  # type: ignore

    index_cli.add_command(hypercorn_cli, name="hypercorn")

try:
    import uvicorn
except ImportError:
    pass
else:
    from .applications import Index

    # See https://stackoverflow.com/questions/58133694/graceful-shutdown-of-uvicorn-starlette-app-with-websockets
    origin_handle_exit = uvicorn.Server.handle_exit

    def handle_exit(self: uvicorn.Server, sig, frame):
        application = self.config.loaded_app
        while not isinstance(application, Index):
            application = application.app
        application.should_exit = True
        return origin_handle_exit(self, sig, frame)

    uvicorn.Server.handle_exit = handle_exit

    @click.command(help="use uvicorn to run Index.py application")
    @click.option(
        "--bind",
        default="127.0.0.1:4190",
        show_default=True,
        help="A string of the form: HOST:PORT, unix:PATH, fd://FD.",
    )
    @click.option("--autoreload/--no-autoreload", default=True, show_default=True)
    @click.option(
        "--log-level",
        type=click.Choice(["critical", "error", "warning", "info", "debug"]),
        default="info",
        show_default=True,
    )
    @click.argument("application")
    def uvicorn_cli(application: str, bind: str, autoreload: bool, log_level: str):
        sys.path.insert(0, os.getcwd())

        if bind.startswith("unix:"):
            bind_config = {"uds": bind[5:] | F(os.path.normpath) | F(os.path.abspath)}
            if autoreload:
                click.secho(
                    "Reload option doesnt work with unix sockets "
                    "in uvicorn: https://github.com/encode/uvicorn/issues/722",
                    fg="yellow",
                )
        elif bind.startswith("fd://"):
            bind_config = {"fd": int(bind[5:])}
            if autoreload:
                click.secho(
                    "Reload option doesnt work with fd "
                    "in uvicorn: https://github.com/encode/uvicorn/issues/368",
                    fg="yellow",
                )
        else:
            if ":" in bind:
                host, port = bind.split(":")
                bind_config = {"host": host, "port": int(port)}
            else:
                bind_config = {"host": bind, "port": 4190}

        uvicorn.run(
            application,
            **bind_config,
            log_level=log_level,
            interface="asgi3",
            lifespan="on",
            reload=autoreload,
        )

    index_cli.add_command(uvicorn_cli, "uvicorn")

try:
    import gunicorn

    assert gunicorn.version_info > (20, 1)
    del gunicorn
except ImportError:
    pass
else:
    MASTER_PID_FILE = ".gunicorn.pid"

    def read_gunicorn_master_pid(pid_file: str = MASTER_PID_FILE) -> int:
        try:
            with open(os.path.join(os.getcwd(), pid_file), "r") as file:
                return int(file.read())
        except FileNotFoundError:
            sys.exit(
                (
                    f'File "{pid_file}" not found, '
                    + "please make sure you have started gunicorn using the "
                    + "`index-cli gunicorn start ...`."
                )
            )

    @click.group(help="use gunicorn to run Index.py application")
    def gunicorn_cli():
        pass

    @gunicorn_cli.command(help="Run gunicorn")
    @click.option(
        "--bind",
        default="127.0.0.1:4190",
        show_default=True,
        help="A string of the form: HOST:PORT, unix:PATH, fd://FD.",
    )
    @click.option("--autoreload/--no-autoreload", default=False, show_default=True)
    @click.option(
        "--log-level",
        type=click.Choice(["critical", "error", "warning", "info", "debug"]),
        default="info",
        show_default=True,
    )
    @click.option("--workers", "-w", default=cpu_count(), show_default=True)
    @click.option(
        "--worker-class",
        "-k",
        default="uvicorn.workers.UvicornWorker",
        show_default=True,
    )
    @click.option("--daemon", "-d", default=False, is_flag=True, show_default=True)
    @click.option(
        "--configuration",
        "-c",
        type=click.Path(exists=True, file_okay=True, dir_okay=False, readable=True),
    )
    @click.argument("application")
    def start(
        workers: int,
        worker_class: str,
        daemon: bool,
        configuration: str,
        application: str,
        bind: str,
        autoreload: bool,
        log_level: str,
    ):
        command = (
            f"{sys.executable} -m gunicorn -k {worker_class}"
            + f" --bind {bind}"
            + f" --chdir {os.getcwd()}"
            + f" --workers {workers}"
            + f" --pid {MASTER_PID_FILE}"
            + f" --log-level {log_level}"
        )
        args = command.split(" ")
        if daemon:
            args.extend("-D --log-file gunicorn.log".split(" "))
        if autoreload:
            args.append("--reload")
        if configuration:
            args.append("-c")
            args.append(configuration.strip())
        args.append(application)

        execute(args)

    # Gunicorn signal handler
    # https://docs.gunicorn.org/en/stable/signals.html

    @gunicorn_cli.command(help="Increment the number of processes by one")
    def incr():
        os.kill(read_gunicorn_master_pid(), signal.SIGTTIN)

    @gunicorn_cli.command(help="Decrement the number of processes by one")
    def decr():
        os.kill(read_gunicorn_master_pid(), signal.SIGTTOU)

    @gunicorn_cli.command(help="Stop gunicorn processes")
    @click.option("--force", "-f", default=False, is_flag=True)
    def stop(force):
        os.kill(read_gunicorn_master_pid(), signal.SIGINT if force else signal.SIGTERM)

    @gunicorn_cli.command(help="Reload configuration and recreate worker processes")
    def reload():
        os.kill(read_gunicorn_master_pid(), signal.SIGHUP)

    @gunicorn_cli.command(help="Restart gunicorn master processes and worker processes")
    @click.option("--force-stop", "-f", default=False, is_flag=True)
    def restart(force_stop):
        oldpid = read_gunicorn_master_pid()

        os.kill(oldpid, signal.SIGUSR2)
        # Waiting for starting new master process and worker processes
        while not os.path.exists(os.path.join(os.getcwd(), MASTER_PID_FILE + ".2")):
            time.sleep(0.5)
        # Stop old master process and worker processes
        os.kill(oldpid, signal.SIGINT if force_stop else signal.SIGTERM)

    index_cli.add_command(gunicorn_cli, "gunicorn")


index_cli.add_command(display_urls, "display-urls")

import_module("commands")



from __future__ import annotations

import operator
import typing
from dataclasses import dataclass
from functools import reduce

from ..parameters import auto_params, update_wrapper

T_Endpoint = typing.Callable[..., typing.Awaitable[typing.Any]]
_RouteSelf = typing.TypeVar("_RouteSelf", bound="BaseRoute")


@dataclass
class BaseRoute:
    path: str
    endpoint: typing.Callable[..., typing.Awaitable[typing.Any]]
    name: typing.Optional[str] = ""

    def extend_middlewares(self, routes: typing.Iterable[BaseRoute]) -> None:
        raise NotImplementedError()

    def _extend_middlewares(
        self, middlewares: typing.Iterable[typing.Callable[[T_Endpoint], T_Endpoint]]
    ) -> None:
        reduce(operator.matmul, middlewares, self)

    def __matmul__(
        self: _RouteSelf, decorator: typing.Callable[[T_Endpoint], T_Endpoint]
    ) -> _RouteSelf:
        endpoint = self.endpoint
        if hasattr(endpoint, "__methods__"):
            for method in map(str.lower, endpoint.__methods__):  # type: ignore
                old_callback = getattr(endpoint, method)
                new_callback = decorator(old_callback)
                if new_callback is not old_callback:
                    update_wrapper(new_callback, old_callback)
                    new_callback = auto_params(new_callback)
                setattr(endpoint, method, new_callback)
        else:
            old_callback = endpoint
            new_callback = decorator(old_callback)
            if new_callback is not old_callback:
                update_wrapper(new_callback, old_callback)
                new_callback = auto_params(new_callback)
            self.endpoint = new_callback
        return self

    def __post_init__(self) -> None:
        assert (
            self.path.startswith("/") or self.path == ""
        ), "Route path must start with '/'"
        if self.name == "":
            self.name = self.endpoint.__name__
        self.endpoint = auto_params(self.endpoint)


@dataclass
class HttpRoute(BaseRoute):
    summary: typing.Optional[str] = None
    description: typing.Optional[str] = None
    tags: typing.Optional[typing.Iterable[str]] = None

    def __post_init__(self) -> None:
        super().__post_init__()
        if self.summary:
            setattr(self.endpoint, "__summary__", self.summary)

        if self.description:
            setattr(self.endpoint, "__description__", self.description)

        if self.tags:
            setattr(self.endpoint, "__tags__", list(self.tags))

    def extend_middlewares(self, routes: typing.Iterable[BaseRoute]) -> None:
        self._extend_middlewares(getattr(routes, "_http_middlewares", []))


@dataclass
class SocketRoute(BaseRoute):
    def extend_middlewares(self, routes: typing.Iterable[BaseRoute]) -> None:
        self._extend_middlewares(getattr(routes, "_socket_middlewares", []))



import importlib
import typing
from functools import reduce, update_wrapper
from pathlib import Path

from ...routing.routes import BaseRoute, HttpRoute, SocketRoute
from ...utils import F


class FileRoutes(typing.Iterable[BaseRoute]):
    def __init__(
        self,
        module_name: str,
        *,
        namespace: str = "",
        allow_underline: bool = False,
        suffix: str = "",
    ) -> None:
        dirpath = Path(importlib.import_module(module_name).__file__).absolute().parent

        self.namespace = namespace
        self._list: typing.List[typing.Union[HttpRoute, SocketRoute]] = []

        for relpath in map(
            lambda pypath: (
                dirpath
                | F(pypath.relative_to)
                | F(str)
                | F(lambda path: path.replace("\\", "/"))
                | F(lambda path: path[:-3])
            ),
            dirpath.glob("**/*.py"),
        ):
            url_path = (
                ("/" + relpath)
                | F(lambda path: (allow_underline or path) and path.replace("_", "-"))
                | F(
                    lambda path: path[:-5] if path.endswith("/index") else path + suffix
                )
            )
            path_list = [module_name, *relpath.split("/")]
            module = importlib.import_module(".".join(path_list))

            append_route = lambda view, middleware_name, route_type: (
                range(len(path_list), 0, -1)
                | F(map, lambda deep: ".".join(path_list[:deep]))
                | F(map, lambda module_name: importlib.import_module(module_name))
                | F(map, lambda module: getattr(module, middleware_name, None))
                | F(filter, lambda middleware: middleware is not None)
                | F(reduce, lambda h, m: update_wrapper(m(h), h), ..., view)
                | F(route_type, url_path, ..., getattr(module, "name", None))
                | F(self._list.append)
            )
            getattr(module, "HTTP", None) | F(
                lambda view: (
                    view is None or append_route(view, "HTTPMiddleware", HttpRoute)
                )
            )
            getattr(module, "Socket", None) | F(
                lambda view: (
                    view is None or append_route(view, "SocketMiddleware", SocketRoute)
                )
            )

    def __iter__(self) -> typing.Iterator[BaseRoute]:
        return iter(self._list)



"""
OpenAPI specification 3.0.3

https://swagger.io/specification/
"""
from __future__ import annotations

import sys
from typing import Any, Dict, List, Union

if sys.version_info[:2] < (3, 8):
    from typing_extensions import Literal
else:
    from typing import Literal

from baize.typing import TypedDict


class _OpenAPIRequired(TypedDict):
    openapi: str
    info: Info
    paths: Paths


class OpenAPI(_OpenAPIRequired, total=False):
    servers: List[Server]
    components: Components
    security: List[SecurityRequirement]
    tags: List[Tag]
    externalDocs: ExternalDocumentation


class _InfoRequired(TypedDict):
    title: str
    version: str


class Info(_InfoRequired, total=False):
    description: str
    terms_of_service: str
    contact: Contact
    license: License


class Contact(TypedDict, total=False):
    name: str
    url: str
    email: str


class _LicenseRequired(TypedDict):
    name: str


class License(_LicenseRequired, total=False):
    url: str


class _ServerRequired(TypedDict):
    url: str


class Server(_ServerRequired, total=False):
    description: str
    variables: Dict[str, ServerVariable]


class _ServerVariableRequired(TypedDict):
    default: str


class ServerVariable(_ServerVariableRequired, total=False):
    enum: List[str]
    description: str


class Components(TypedDict, total=False):
    schemas: Dict[str, Schema | Reference]
    responses: Dict[str, Response | Reference]
    parameters: Dict[str, Parameter | Reference]
    examples: Dict[str, Example | Reference]
    requestBodies: Dict[str, RequestBody | Reference]
    headers: Dict[str, Header | Reference]
    securitySchemes: Dict[str, SecurityScheme | Reference]
    links: Dict[str, Link | Reference]
    callbacks: Dict[str, Callback | Reference]


Paths = Dict[str, "PathItem"]

_PathRef = TypedDict("_PathRef", {"$ref": str}, total=False)


class PathItem(_PathRef, total=False):
    summary: str
    description: str
    get: Operation
    put: Operation
    post: Operation
    delete: Operation
    options: Operation
    head: Operation
    patch: Operation
    trace: Operation
    servers: List[Server]
    parameters: List[Parameter | Reference]


class _OperationRequired(TypedDict):
    responses: Responses


class Operation(_OperationRequired, total=False):
    tags: List[str]
    summary: str
    description: str
    externalDocs: ExternalDocumentation
    operationId: str
    parameters: List[Parameter | Reference]
    requestBody: RequestBody | Reference
    callbacks: Dict[str, Callback | Reference]
    deprecated: bool
    security: List[SecurityRequirement]
    servers: List[Server]


class _ExternalDocumentationRequired(TypedDict):
    url: str


class ExternalDocumentation(_ExternalDocumentationRequired, total=False):
    description: str


# Beacuse "in" is a key word in python
_ParameterRequired = TypedDict("_ParameterRequired", {"name": str, "in": str})


class _ParameterOptional(TypedDict, total=False):
    description: str
    required: bool
    deprecated: bool
    allowEmptyValue: bool

    style: str
    explode: bool
    allowReserved: bool
    schema: Schema | Reference
    example: Any
    examples: Dict[str, Example | Reference]

    content: Dict[str, MediaType]


class Parameter(_ParameterRequired, _ParameterOptional):
    pass


class _RequestBodyRequired(TypedDict):
    content: Dict[str, MediaType]


class RequestBody(_RequestBodyRequired, total=False):
    description: str
    required: bool


class MediaType(TypedDict, total=False):
    schema: Schema | Reference
    example: Any
    examples: Dict[str, Example | Reference]
    encoding: Dict[str, Encoding]


class Encoding(TypedDict, total=False):
    contentType: str
    headers: Dict[str, Header | Reference]
    style: str
    explode: bool
    allowReserved: bool


Responses = Dict[str, "Response"]


class _ResponseRequired(TypedDict):
    description: str


class Response(_ResponseRequired, total=False):
    headers: Dict[str, Header | Reference]
    content: Dict[str, MediaType]
    links: Dict[str, Link | Reference]


Callback = Dict[str, "PathItem"]


class Example(TypedDict, total=False):
    summary: str
    description: str
    value: Any
    externalValue: str


class Link(TypedDict, total=False):
    operationRef: str
    operationId: str
    parameters: Dict[str, Any]
    requestBody: Any
    description: str
    server: Server


class Header(TypedDict, total=False):
    description: str
    required: bool
    deprecated: bool
    allowEmptyValue: bool

    style: str
    explode: bool
    allowReserved: bool
    schema: Schema | Reference
    example: Any
    examples: Dict[str, Example | Reference]

    content: Dict[str, MediaType]


class _TagRequired(TypedDict):
    name: str


class Tag(_TagRequired, total=False):
    description: str
    external_docs: ExternalDocumentation


Reference = TypedDict("Reference", {"$ref": str})


class Schema(TypedDict, total=False):
    nullable: bool
    discriminator: Discriminator
    readOnly: bool
    writeOnly: bool
    xml: XML
    externalDocs: ExternalDocumentation
    example: Any
    deprecated: bool


class _DiscriminatorRequired(TypedDict):
    propertyName: str


class Discriminator(_DiscriminatorRequired, total=False):
    mapping: Dict[str, str]


class XML(TypedDict, total=False):
    name: str
    namespace: str
    prefix: str
    attribute: str
    wrapped: bool


# Beacuse "in" is a key word in python
_ApiKeyScheme = TypedDict(
    "_ApiKeyScheme",
    {
        "type": Literal["apiKey"],
        "name": str,
        "in": Literal["query", "header", "cookie"],
    },
)


class _SecuritySchemeOptional(TypedDict, total=False):
    description: str


class ApiKeyScheme(_ApiKeyScheme, _SecuritySchemeOptional):
    pass


class HttpScheme(_SecuritySchemeOptional):
    type: Literal["http"]
    scheme: str


class _HttpBearerSchemeOptional(TypedDict, total=False):
    bearerFormat: str


class HttpBearerScheme(HttpScheme, _HttpBearerSchemeOptional):
    scheme: Literal["bearer"]


class OAuth2Scheme(_SecuritySchemeOptional):
    type: Literal["oauth2"]
    flows: OAuthFlows


class OpenIdConnectScheme(_SecuritySchemeOptional):
    type: Literal["openIdConnect"]
    openIdConnectUrl: str


SecurityScheme = Union[
    ApiKeyScheme, HttpScheme, HttpBearerScheme, OAuth2Scheme, OpenIdConnectScheme
]


class OAuthFlows(TypedDict, total=False):
    implicit: OAuthFlowImplicit
    password: OAuthFlowPassword
    clientCredentials: OAuthFlowClientCredentials
    authorizationCode: OAuthFlowAuthorizationCode


class _OAuthFlowBase(TypedDict, total=False):
    refreshUrl: str


class OAuthFlowBase(_OAuthFlowBase):
    scopes: Dict[str, str]


class OAuthFlowImplicit(OAuthFlowBase):
    authorizationUrl: str


class OAuthFlowPassword(OAuthFlowBase):
    tokenUrl: str


class OAuthFlowClientCredentials(OAuthFlowBase):
    tokenUrl: str


class OAuthFlowAuthorizationCode(OAuthFlowBase):
    authorizationUrl: str
    tokenUrl: str


SecurityRequirement = Dict[str, List[str]]



from __future__ import annotations

from abc import ABCMeta, abstractmethod
from typing import Any, List, Mapping, Union

from baize.asgi import SmallResponse

from .requests import request
from .responses import HttpResponse


class BaseTemplates(metaclass=ABCMeta):
    @abstractmethod
    def TemplateResponse(
        self,
        name: str,
        context: Mapping[str, Any],
        status_code: int = 200,
        headers: Mapping[str, str] = None,
        media_type: str = None,
        charset: str = None,
    ) -> HttpResponse:
        """
        The subclass must override this method and return
        an instance of a Response object.
        """


try:
    import jinja2
except ImportError:
    pass
else:

    class _Jinja2TemplateResponse(SmallResponse):
        media_type = "text/html"

        def __init__(
            self,
            env: jinja2.Environment,
            name: str,
            context: Mapping[str, Any],
            status_code: int = 200,
            headers: Mapping[str, str] = None,
            media_type: str = None,
            charset: str = None,
        ):
            self.env = env
            self.template = self.env.get_template(name)
            super().__init__(context, status_code, headers, media_type, charset)

        async def render(self, context: Mapping[str, Any]) -> bytes:
            if self.env.enable_async:  # type: ignore
                text = await self.template.render_async(context)
            else:
                text = self.template.render(context)
            return text.encode(self.charset)

    class Jinja2Templates(BaseTemplates):
        """
        templates = Jinja2Templates("dir", "package:dir")

        return templates.TemplateResponse("index.html", {"request": request})
        """

        def __init__(self, *directories: str) -> None:
            self.env = self.get_env(self.get_loaders(*directories))

        def get_loaders(self, *directories: str) -> jinja2.BaseLoader:
            templates_loaders: List[
                Union[jinja2.FileSystemLoader, jinja2.PackageLoader]
            ] = []
            for directory in directories:
                if ":" in directory:
                    package_name, package_path = directory.split(":", maxsplit=1)
                    templates_loaders.append(
                        jinja2.PackageLoader(package_name, package_path)
                    )
                else:
                    templates_loaders.append(jinja2.FileSystemLoader(directory))
            return jinja2.ChoiceLoader(templates_loaders)

        def get_env(self, loader: jinja2.BaseLoader) -> jinja2.Environment:
            @jinja2.contextfunction
            def url_for(context: dict, name: str, **path_params: Any) -> str:
                router = request.app.router
                return router.url_for(name, path_params)

            env = jinja2.Environment(loader=loader, enable_async=True, autoescape=True)
            env.globals["url_for"] = url_for
            return env

        def TemplateResponse(
            self,
            name: str,
            context: Mapping[str, Any],
            status_code: int = 200,
            headers: Mapping[str, str] = None,
            media_type: str = None,
            charset: str = None,
        ) -> _Jinja2TemplateResponse:
            return _Jinja2TemplateResponse(
                self.env,
                name,
                context,
                status_code=status_code,
                headers=headers,
                media_type=media_type,
                charset=charset,
            )



from __future__ import annotations

import typing
from http import HTTPStatus

from baize.asgi import FileResponse, HTMLResponse
from baize.asgi import JSONResponse as _JSONResponse
from baize.asgi import PlainTextResponse, RedirectResponse
from baize.asgi import Response as HttpResponse
from baize.asgi import SendEventResponse, ServerSentEvent, StreamResponse
from pydantic import BaseModel, create_model
from pydantic.json import pydantic_encoder
from pydantic.typing import display_as_type

from .requests import request
from .utils import safe_issubclass

__all__ = [
    "convert_response",
    "HttpResponse",
    "FileResponse",
    "HTMLResponse",
    "JSONResponse",
    "PlainTextResponse",
    "RedirectResponse",
    "SendEventResponse",
    "ServerSentEvent",
    "StreamResponse",
    "TemplateResponse",
]


class JSONResponse(_JSONResponse):
    def __init__(
        self,
        content: typing.Any,
        status_code: int = 200,
        headers: typing.Mapping[str, str] = None,
        **kwargs: typing.Any,
    ) -> None:
        json_kwargs: typing.Dict[str, typing.Any] = {
            "default": pydantic_encoder,
        }
        json_kwargs.update(**kwargs)
        super().__init__(
            content, status_code=status_code, headers=headers, **json_kwargs
        )


def TemplateResponse(
    name: str,
    context: typing.Mapping[str, typing.Any],
    status_code: int = 200,
    headers: typing.Mapping[str, str] = None,
    media_type: str = None,
    charset: str = None,
) -> HttpResponse:
    templates = request.app.templates
    if templates is None:
        raise RuntimeError(
            "You must assign a value to `app.templates` to use TemplateResponse"
        )

    return templates.TemplateResponse(
        name, context, status_code, headers, media_type, charset
    )


def convert_response(response: typing.Any) -> HttpResponse:
    """
    shortcut for automatic

    Example:

        response = convert(response)

    It is equivalent to:

        if isinstance(response, tuple):
            response = automatic(*response)
        else:
            response = automatic(response)

    """
    if isinstance(response, tuple):
        return request.app.response_converter(*response)
    else:
        return request.app.response_converter(response)


def http_response__class_getitem__(parameters):
    """
    Use HttpResponse[status, headers, content] to describe response
    """
    if isinstance(parameters, tuple):
        assert len(parameters) in (2, 3)
        if len(parameters) == 2:
            (status_code, headers), content = parameters, {}
        else:
            status_code, headers, content = parameters
    else:
        status_code, headers, content = parameters, {}, {}
    assert isinstance(status_code, int) or status_code == "default"

    docs = {
        str(status_code): {
            "description": HTTPStatus(status_code).description,
        }
    }
    if headers:
        docs[str(status_code)]["headers"] = headers

    if content:
        docs[str(status_code)]["content"] = content

    return docs


HttpResponse.__class_getitem__ = http_response__class_getitem__  # type: ignore


def html_response__class_getitem__(parameters):
    """
    Use HTMLResponse[status, headers] to describe response
    """
    if isinstance(parameters, tuple):
        status_code, headers = parameters
    else:
        status_code, headers = parameters, {}
    assert isinstance(status_code, int)
    docs = {
        str(status_code): {
            "description": HTTPStatus(status_code).description,
            "content": {"text/html": {"schema": {"type": "string"}}},
        }
    }
    if headers:
        docs[str(status_code)]["headers"] = headers
    return docs


HTMLResponse.__class_getitem__ = html_response__class_getitem__  # type: ignore


def plain_text_response__class_getitem__(parameters):
    """
    Use PlainTextResponse[status, headers] to describe response
    """
    if isinstance(parameters, tuple):
        status_code, headers = parameters
    else:
        status_code, headers = parameters, {}
    assert isinstance(status_code, int)
    docs = {
        str(status_code): {
            "description": HTTPStatus(status_code).description,
            "content": {"text/plain": {"schema": {"type": "string"}}},
        }
    }
    if headers:
        docs[str(status_code)]["headers"] = headers
    return docs


PlainTextResponse.__class_getitem__ = plain_text_response__class_getitem__  # type: ignore


def redirect_response__class_getitem__(parameters):
    """
    Use RedirectResponse[status] to describe response
    """
    if isinstance(parameters, tuple):
        status_code, headers = parameters
    else:
        status_code, headers = parameters, {}
    assert isinstance(status_code, int)
    docs = {
        str(status_code): {
            "description": HTTPStatus(status_code).description,
            "headers": {
                "Location": {"schema": {"type": "string"}},
            },
        }
    }
    if headers:
        docs[str(status_code)]["headers"].update(headers)
    return docs


RedirectResponse.__class_getitem__ = redirect_response__class_getitem__  # type: ignore


def file_response__class_getitem__(parameters):
    """
    Use FileResponse[content_type, headers] to describe response
    """
    if isinstance(parameters, tuple):
        content_type, headers = parameters
    else:
        content_type, headers = parameters, {}
    assert isinstance(content_type, str)
    docs = {
        "200": {
            "description": HTTPStatus.OK.description,
            "content": {
                content_type: {"schema": {"type": "string", "format": "binary"}}
            },
        },
        "206": {
            "description": HTTPStatus.PARTIAL_CONTENT.description,
            "content": {
                content_type: {"schema": {"type": "string", "format": "binary"}}
            },
        },
    }
    if headers:
        docs["200"]["headers"] = headers
        docs["206"]["headers"] = headers
    return docs


FileResponse.__class_getitem__ = file_response__class_getitem__  # type: ignore


def json_response__class_getitem__(parameters):
    """
    Use JSONResponse[status, headers, content] to describe response
    """
    if isinstance(parameters, tuple):
        assert len(parameters) in (2, 3)
        if len(parameters) == 2:
            (status_code, headers), content = parameters, {}
        else:
            status_code, headers, content = parameters
    else:
        status_code, headers, content = parameters, {}, {}
    assert isinstance(status_code, int) or status_code == "default"

    docs = {
        str(status_code): {
            "description": HTTPStatus(status_code).description,
        }
    }
    if headers:
        docs[str(status_code)]["headers"] = headers

    if content:
        if isinstance(content, dict) or (
            getattr(content, "__origin__", None) is None
            and safe_issubclass(content, BaseModel)
        ):
            real_content = content
        else:
            real_content = create_model(
                f"ParsingModel[{display_as_type(content)}]", __root__=(content, ...)
            )
        docs[str(status_code)]["content"] = real_content

    return docs


JSONResponse.__class_getitem__ = json_response__class_getitem__  # type: ignore



from __future__ import annotations

import abc
import operator
import sys
import typing
from copy import deepcopy
from functools import reduce

if sys.version_info[:2] < (3, 8):
    from typing_extensions import Literal
else:
    from typing import Literal

from baize.routing import compile_path

from ..utils import FF, F
from ..views import required_method
from .routes import BaseRoute, HttpRoute, SocketRoute
from .tree import RadixTree, RouteType


class NoMatchFound(Exception):
    """
    Raised by `.search(path)` if no matching route exists.
    """


class NoRouteFound(Exception):
    """
    Raised by `.url_for(name, **path_params)` if no matching route exists.
    """


T = typing.TypeVar("T")

_RRMixinSelf = typing.TypeVar("_RRMixinSelf", bound="RouteRegisterMixin")
View = typing.TypeVar("View", bound=typing.Callable)


class HttpRegister:
    def __init__(self, routes: RouteRegisterMixin) -> None:
        self.__routes = routes

    def _register_with_method(
        self,
        method: str,
        path: str,
        *,
        name: str = "",
        middlewares: typing.Iterable[typing.Callable[[View], View]] = [],
        summary: str = None,
        description: str = None,
        tags: typing.Iterable[str] = None,
    ) -> typing.Callable[[View], View]:
        """
        if method == "any", all request method would be allowed.
        """

        def register(endpoint: View) -> View:
            route: HttpRoute = reduce(
                operator.matmul,
                middlewares,
                HttpRoute(path, endpoint, name, summary, description, tags),
            )
            if method != "any":
                route = route @ required_method(method.upper())

            self.__routes << route
            return endpoint

        return register

    def __call__(
        self,
        path: str,
        *,
        name: str = "",
        middlewares: typing.Iterable[typing.Callable[[View], View]] = [],
        summary: str = None,
        description: str = None,
        tags: typing.Iterable[str] = None,
    ) -> typing.Callable[[View], View]:
        """
        shortcut for `self << HttpRoute(path, endpoint, name)`

        example:
        ```python
            @routes.http("/path", name="endpoint-name")
            class Endpoint(HttpView): ...
        ```
        """
        return self._register_with_method(
            "any",
            path,
            name=name,
            middlewares=middlewares,
            summary=summary,
            description=description,
            tags=tags,
        )

    def get(
        self,
        path: str,
        *,
        name: str = "",
        middlewares: typing.Iterable[typing.Callable[[View], View]] = [],
        summary: str = None,
        description: str = None,
        tags: typing.Iterable[str] = None,
    ) -> typing.Callable[[View], View]:
        """
        shortcut for `self << HttpRoute(path, endpoint, name) @ required_method("GET")`

        example:
        ```python
            @routes.http.get("/path", name="endpoint-name")
            class Endpoint(HttpView): ...
        ```
        """
        return self._register_with_method(
            "get",
            path,
            name=name,
            middlewares=middlewares,
            summary=summary,
            description=description,
            tags=tags,
        )

    def post(
        self,
        path: str,
        *,
        name: str = "",
        middlewares: typing.Iterable[typing.Callable[[View], View]] = [],
        summary: str = None,
        description: str = None,
        tags: typing.Iterable[str] = None,
    ) -> typing.Callable[[View], View]:
        """
        shortcut for `self << HttpRoute(path, endpoint, name) @ required_method("POST")`

        example:
        ```python
            @routes.http.post("/path", name="endpoint-name")
            class Endpoint(HttpView): ...
        ```
        """
        return self._register_with_method(
            "post",
            path,
            name=name,
            middlewares=middlewares,
            summary=summary,
            description=description,
            tags=tags,
        )

    def put(
        self,
        path: str,
        *,
        name: str = "",
        middlewares: typing.Iterable[typing.Callable[[View], View]] = [],
        summary: str = None,
        description: str = None,
        tags: typing.Iterable[str] = None,
    ) -> typing.Callable[[View], View]:
        """
        shortcut for `self << HttpRoute(path, endpoint, name) @ required_method("PUT")`

        example:
        ```python
            @routes.http.put("/path", name="endpoint-name")
            class Endpoint(HttpView): ...
        ```
        """
        return self._register_with_method(
            "put",
            path,
            name=name,
            middlewares=middlewares,
            summary=summary,
            description=description,
            tags=tags,
        )

    def patch(
        self,
        path: str,
        *,
        name: str = "",
        middlewares: typing.Iterable[typing.Callable[[View], View]] = [],
        summary: str = None,
        description: str = None,
        tags: typing.Iterable[str] = None,
    ) -> typing.Callable[[View], View]:
        """
        shortcut for `self << HttpRoute(path, endpoint, name) @ required_method("PATCH")`

        example:
        ```python
            @routes.http.patch("/path", name="endpoint-name")
            class Endpoint(HttpView): ...
        ```
        """
        return self._register_with_method(
            "patch",
            path,
            name=name,
            middlewares=middlewares,
            summary=summary,
            description=description,
            tags=tags,
        )

    def delete(
        self,
        path: str,
        *,
        name: str = "",
        middlewares: typing.Iterable[typing.Callable[[View], View]] = [],
        summary: str = None,
        description: str = None,
        tags: typing.Iterable[str] = None,
    ) -> typing.Callable[[View], View]:
        """
        shortcut for `self << HttpRoute(path, endpoint, name) @ required_method("DELETE")`

        example:
        ```python
            @routes.http.delete("/path", name="endpoint-name")
            class Endpoint(HttpView): ...
        ```
        """
        return self._register_with_method(
            "delete",
            path,
            name=name,
            middlewares=middlewares,
            summary=summary,
            description=description,
            tags=tags,
        )


class RouteRegisterMixin(abc.ABC):
    @abc.abstractmethod
    def append(self: _RRMixinSelf, route: BaseRoute) -> _RRMixinSelf:
        raise NotImplementedError

    def __lshift__(
        self: _RRMixinSelf, other: typing.Union[BaseRoute, typing.Iterable[BaseRoute]]
    ) -> _RRMixinSelf:
        """
        self << routes
        """
        if isinstance(other, BaseRoute):
            return self.append(other)
        elif isinstance(other, typing.Iterable):
            for route in other:
                if isinstance(route, BaseRoute):
                    if getattr(other, "namespace", "") and route.name:
                        route.name = getattr(other, "namespace") + ":" + route.name
                    route.extend_middlewares(other)
                self << route
            return self
        else:
            return NotImplemented

    @property
    def http(self) -> HttpRegister:
        return HttpRegister(self)

    def websocket(
        self,
        path: str,
        *,
        name: str = "",
        middlewares: typing.Iterable[typing.Callable[[View], View]] = [],
    ) -> typing.Callable[[View], View]:
        """
        shortcut for `self << SocketRoute(path, endpoint, name)`

        example:
        ```python
            @routes.websocket("/path", name="endpoint-name")
            class Endpoint(SocketView): ...
        ```
        """

        def register(endpoint: View) -> View:
            self << reduce(
                operator.matmul, middlewares, SocketRoute(path, endpoint, name)
            )
            return endpoint

        return register


_RoutesSelf = typing.TypeVar("_RoutesSelf", bound="Routes")


class Routes(typing.Sequence[BaseRoute], RouteRegisterMixin):
    def __init__(
        self,
        *iterable: typing.Union[BaseRoute, typing.Iterable[BaseRoute]],
        namespace: str = "",
        tags: typing.Iterable[str] = None,
        http_middlewares: typing.Sequence[typing.Any] = [],
        socket_middlewares: typing.Sequence[typing.Any] = [],
    ) -> None:
        self.namespace = namespace
        self._list: typing.List[BaseRoute] = []
        self._http_middlewares = list(http_middlewares)
        self._http_middlewares.append(
            lambda endpoint: (
                setattr(  # type: ignore
                    endpoint,
                    "__tags__",
                    list(getattr(endpoint, "__tags__", [])) + list(tags or []),
                )
                or endpoint
            )
        )
        self._socket_middlewares = list(socket_middlewares)
        for route in iterable:
            self << route

    @typing.overload
    def __getitem__(self, index: int) -> BaseRoute:
        ...

    @typing.overload
    def __getitem__(self, index: slice) -> typing.NoReturn:
        ...

    def __getitem__(self, index):
        if isinstance(index, int):
            return self._list[index]
        else:
            raise TypeError("Slicing syntax is not allowed")

    def __len__(self) -> int:
        return len(self._list)

    def append(self: _RoutesSelf, route: BaseRoute) -> _RoutesSelf:
        self._list.append(route)
        return self

    def __rfloordiv__(self: _RoutesSelf, other: str) -> _RoutesSelf:
        """
        other // self
        """
        if not isinstance(other, str):
            return NotImplemented

        return Prefix(other) // self

    def __add__(self, routes: typing.Iterable[BaseRoute]) -> Routes:
        """
        self + routes
        """
        return Routes() << self << routes

    def __radd__(self, routes: typing.Iterable[BaseRoute]) -> Routes:
        """
        routes + self
        """
        return Routes() << routes << self

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, typing.Sequence):
            return NotImplemented
        return len(o) == len(self) and all(
            zip(self, o) | F(map, FF(lambda r, r_: r == r_))
        )

    def http_middleware(self, middleware: T) -> T:
        """
        append middleware in routes

        example:
        ```
            @routes.http_middleware
            def middleware(endpoint):
                async def wrapper():
                    return await endpoint()
                return wrapper
        ```
        """
        self._http_middlewares.append(middleware)
        return middleware

    def socket_middleware(self, middleware: T) -> T:
        """
        append middleware in routes

        example:
        ```
            @routes.socket_middleware
            def middleware(endpoint):
                async def wrapper():
                    await endpoint()
                return wrapper
        ```
        """
        self._socket_middlewares.append(middleware)
        return middleware


_RouteSequence = typing.TypeVar("_RouteSequence", bound=typing.Iterable[BaseRoute])


class Prefix(str):
    def __init__(self, *args, **kwargs) -> None:
        assert self.startswith("/") and not self.endswith("/")

    def __floordiv__(self, other: _RouteSequence) -> _RouteSequence:
        """
        self // other
        """
        if not isinstance(other, typing.Iterable):
            return NotImplemented
        result = deepcopy(other)
        for route in result:
            route.path = self + route.path
        return typing.cast(_RouteSequence, result)


_RouterSelf = typing.TypeVar("_RouterSelf", bound="Router")


class Router(RouteRegisterMixin):
    def __init__(self, routes: typing.Iterable[BaseRoute]) -> None:
        self.http_tree = RadixTree()
        self.websocket_tree = RadixTree()

        self.http_routes: typing.Dict[str, RouteType] = {}
        self.websocket_routes: typing.Dict[str, RouteType] = {}

        self << routes

    def append(self: _RouterSelf, route: BaseRoute) -> _RouterSelf:
        if isinstance(route, HttpRoute):
            radix_tree = self.http_tree
            routes = self.http_routes
        elif isinstance(route, SocketRoute):
            radix_tree = self.websocket_tree
            routes = self.websocket_routes
        else:
            raise TypeError(
                f"Need type: `HttpRoute` or `SocketRoute`, but got type: {type(route)}"
            )

        if route.path == "":
            route.path = "/"

        if route.name in routes:
            raise ValueError(f"Duplicate route name: {route.name}")

        radix_tree.append(route.path, route.endpoint)
        path_format, path_convertors = compile_path(route.path)

        if route.name:  # name not in ("", None)
            routes[route.name] = (path_format, path_convertors, route.endpoint)

        return self

    def search(
        self, protocol: Literal["http", "websocket"], path: str
    ) -> typing.Tuple[typing.Dict[str, typing.Any], typing.Callable[[], typing.Any]]:
        if protocol == "http":
            radix_tree = self.http_tree
        elif protocol == "websocket":
            radix_tree = self.websocket_tree
        else:
            raise ValueError("`protocol` must be in ('http', 'websocket')")

        params, endpoint = radix_tree.search(path)

        if params is None or endpoint is None:
            raise NoMatchFound(path)

        return params, endpoint

    def url_for(
        self,
        name: str,
        path_params: typing.Dict[str, typing.Any] = {},
        *,
        protocol: Literal["http", "websocket"] = "http",
    ) -> str:
        if protocol == "http":
            routes = self.http_routes
        elif protocol == "websocket":
            routes = self.websocket_routes
        else:
            raise ValueError("`protocol` must be in ('http', 'websocket')")

        if name not in routes:
            raise NoRouteFound(f"No route with name '{name}' exists")

        path_format, path_convertors, _ = routes[name]

        return path_format.format_map(
            {
                name: path_convertors[name].to_string(value)
                for name, value in path_params.items()
            }
        )



import typing
from functools import wraps

from ...routing import BaseRoute, HttpRoute, Routes
from ...views import HttpView

_RoutesSelf = typing.TypeVar("_RoutesSelf", bound="MultimethodRoutes")

exclude_self = lambda func: wraps(func)(
    lambda self, *args, **kwargs: func(*args, **kwargs)
)


class MultimethodRoutes(Routes):
    def __init__(
        self,
        *iterable: typing.Union[BaseRoute, typing.Iterable[BaseRoute]],
        namespace: str = "",
        tags: typing.Iterable[str] = None,
        http_middlewares: typing.Sequence[typing.Any] = [],
        socket_middlewares: typing.Sequence[typing.Any] = [],
        base_class: typing.Type[HttpView] = HttpView,
    ) -> None:
        super().__init__(
            *iterable,
            namespace=namespace,
            tags=tags,
            http_middlewares=http_middlewares,
            socket_middlewares=socket_middlewares,
        )
        self.base_class = base_class

    def append(self: _RoutesSelf, route: BaseRoute) -> _RoutesSelf:
        if hasattr(route.endpoint, "__methods__"):
            raise TypeError("MultimethodRoutes not allow use class-base view.")

        if not isinstance(route, HttpRoute):
            self._list.append(route)
            return self

        try:
            r = next(
                filter(
                    lambda r: isinstance(r, HttpRoute) and r.path == route.path,
                    self._list,
                )
            )
            if not hasattr(route.endpoint, "__method__") or not (
                hasattr(r.endpoint, "__method__") or hasattr(r.endpoint, "__methods__")
            ):
                raise RuntimeError(
                    f"Routing '{route.path}' conflict, can be resolved by restricting the request method."
                )
        except StopIteration:
            self._list.append(route)
        else:
            if hasattr(r.endpoint, "__methods__"):
                endpoint = type(
                    r.endpoint.__name__,
                    (self.base_class, _MultiMethodView),
                    {
                        **{
                            method.lower(): getattr(r.endpoint, method.lower())
                            for method in r.endpoint.__methods__  # type: ignore
                        },
                        route.endpoint.__method__.lower(): exclude_self(route.endpoint),  # type: ignore
                    },
                )
            else:
                endpoint = type(
                    "_MultimethodEndpoint",
                    (self.base_class, _MultiMethodView),
                    {
                        r.endpoint.__method__.lower(): exclude_self(r.endpoint),  # type: ignore
                        route.endpoint.__method__.lower(): exclude_self(route.endpoint),  # type: ignore
                    },
                )
            # replacing route inplace
            r.endpoint = endpoint
        return self


class _MultiMethodView:
    """
    Just as a mark
    """


def is_multimethod_view(cls: type) -> bool:
    return isinstance(cls, type) and issubclass(cls, _MultiMethodView)



from .routers import NoMatchFound, NoRouteFound, Prefix, Router, Routes
from .routes import BaseRoute, HttpRoute, SocketRoute

__all__ = [
    "Router",
    "Routes",
    "BaseRoute",
    "HttpRoute",
    "SocketRoute",
    "Prefix",
    "NoMatchFound",
    "NoRouteFound",
]



from __future__ import annotations

import sys
from dataclasses import dataclass
from typing import Any, Callable, Optional

if sys.version_info[:2] < (3, 8):
    from typing_extensions import Literal
else:
    from typing import Literal

from pydantic.fields import FieldInfo as _FieldInfo
from pydantic.fields import NoArgAnyCallable, Undefined


class FieldInfo(_FieldInfo):
    __slots__ = _FieldInfo.__slots__

    _in: Literal["path", "query", "header", "cookie", "body", "request"]


class PathInfo(FieldInfo):
    __slots__ = ("exclusive", *FieldInfo.__slots__)

    _in: Literal["path"] = "path"

    def __init__(self, default: Any = Undefined, **kwargs: Any) -> None:
        self.exclusive = kwargs.pop("exclusive")
        super().__init__(default, **kwargs)


class QueryInfo(FieldInfo):
    __slots__ = ("exclusive", *FieldInfo.__slots__)

    _in: Literal["query"] = "query"

    def __init__(self, default: Any = Undefined, **kwargs: Any) -> None:
        self.exclusive = kwargs.pop("exclusive")
        super().__init__(default, **kwargs)


class HeaderInfo(FieldInfo):
    __slots__ = ("exclusive", *FieldInfo.__slots__)

    _in: Literal["header"] = "header"

    def __init__(self, default: Any = Undefined, **kwargs: Any) -> None:
        self.exclusive = kwargs.pop("exclusive")
        super().__init__(default, **kwargs)


class CookieInfo(FieldInfo):
    __slots__ = ("exclusive", *FieldInfo.__slots__)

    _in: Literal["cookie"] = "cookie"

    def __init__(self, default: Any = Undefined, **kwargs: Any) -> None:
        self.exclusive = kwargs.pop("exclusive")
        super().__init__(default, **kwargs)


class BodyInfo(FieldInfo):
    __slots__ = ("exclusive", *FieldInfo.__slots__)

    _in: Literal["body"] = "body"

    def __init__(self, default: Any = Undefined, **kwargs: Any) -> None:
        self.exclusive = kwargs.pop("exclusive")
        super().__init__(default, **kwargs)


@dataclass
class RequestInfo:
    __slots__ = ("default", "default_factory", "alias")

    default: Any
    default_factory: Optional[NoArgAnyCallable]
    alias: Optional[str]


@dataclass
class DependInfo:
    __slots__ = ("call", "to_async")

    call: Callable
    to_async: bool



import asyncio
import threading
from typing import Any


class State(dict):
    """
    An object that can be used to store arbitrary state.
    """

    def __enter__(self):
        if not hasattr(self, "sync_lock"):
            self.sync_lock = threading.Lock()
        self.sync_lock.acquire()
        return self

    def __exit__(self, exc_type, value, traceback):
        self.sync_lock.release()

    async def __aenter__(self):
        if not hasattr(self, "async_lock"):
            self.async_lock = asyncio.Lock()
        await self.async_lock.acquire()
        return self

    async def __aexit__(self, exc_type, value, traceback):
        self.async_lock.release()

    def __setattr__(self, name: Any, value: Any) -> None:
        self[name] = value

    def __getattr__(self, name: Any) -> Any:
        try:
            return self[name]
        except KeyError:
            message = "'{}' object has no attribute '{}'"
            raise AttributeError(message.format(self.__class__.__name__, name))

    def __delattr__(self, name: Any) -> None:
        del self[name]
